/***********************************************************\
|                                                           |
| Alien Breed IV: Final Assault - Levelset Editor           |
| Version 1.0                                               |
|                                                           |
| Copyright (c)2001 TI-Calculator Programming Alliance      |
| Written and Programmed by James Vernon <james@calc.org>   |
| ICQ#: 71589304                                            |
| http://jamesv.calc.org                                    |
| http://tcpa.calc.org                                      |
|                                                           |
| You are allowed to use any of this code as long as you    |
| give credit where due.                                    |
|                                                           |
| Compiled using Borland C++ v4.02                          |
|                                                           |
\***********************************************************/

/////////////////////////////////////////////////
// Include Files
/////////////////////////////////////////////////
#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <string.h>
#include <time.h>

/////////////////////////////////////////////////
// Function Prototypes
/////////////////////////////////////////////////
void createNew(void);
void loadLevelset(FILE *fp);
void saveLevelset(char *name);
void editLevels(void);
void helpScreen(void);
void showMap(int levelNo);
void setPasscode(int levelNo);
void highlightBlock(int x, int y);
void highlightTile(int tile, int color);
void showTiles(void);
void showLevel(int l, int xOffset, int yOffset);
int checkAHandSG(void);
void waitKey(void);
void gfxMode(void);
void textMode(void);
void setCursor(unsigned char y, unsigned char x);
void putPixel(int x, int y, int c);
void putBWSprite(int x, int y, unsigned char *spr);
void fillBox(int y, int x, int h, int w, int c);
void line(unsigned short x1, unsigned short y1, unsigned short x2, unsigned short y2, unsigned char c);
void drawBox(int y, int x, int h, int w, int c);
void clearVideo(void);

/////////////////////////////////////////////////
// Keypress Defines
/////////////////////////////////////////////////
#define K_TAB 9
#define K_ENTER 13
#define K_CTRLA 1
#define K_CTRLS 19
#define K_ESC 27
#define K_SPACE ' '
#define K_1 '1'
#define K_2 '2'
#define K_3 '3'
#define K_4 '4'
#define K_5 '5'
#define K_6 '6'
#define K_7 '7'
#define K_8 '8'
#define K_9 '9'
#define K_0 '0'
#define K_Q 'q'
#define K_W 'w'
#define K_E 'e'
#define K_R 'r'
#define K_T 't'
#define K_Y 'y'
#define K_U 'u'
#define K_I 'i'
#define K_O 'o'
#define K_P 'p'
#define K_F1 59
#define K_F2 60
#define K_F3 61
#define K_F4 62
#define K_F5 63
#define K_F6 64
#define K_F7 65
#define K_F8 66
#define K_F9 67
#define K_F10 68
#define K_HOME 71
#define K_UP 72
#define K_LEFT 75
#define K_RIGHT 77
#define K_END 79
#define K_DOWN 80
#define K_PGDOWN 81
#define K_DELETE 83
#define K_CTRLF1 94
#define K_CTRLF2 95
#define K_CTRLF3 96
#define K_CTRLF4 97
#define K_CTRLF5 98
#define K_CTRLF6 99
#define K_CTRLF7 100
#define K_CTRLF8 101
#define K_CTRLF9 102
#define K_CTRLF10 103

/////////////////////////////////////////////////
// Other Assorted Defines
/////////////////////////////////////////////////
#define NO_ERR 0
#define ERR_TILES 1
#define ERR_USAGE 2
#define ERR_INVALIDNAME 3
#define MAX_LEVELS 8
#define MIN_HEIGHT 8
#define MAX_HEIGHT 32
#define MIN_WIDTH 12
#define MAX_WIDTH 32
#define DETECT_LENGTH 8
#define SHOW_TILES_Y 120
#define VIEW_BOX_X 111
#define VIEW_BOX_Y 50
#define BLOCK_BOX_X 290
#define BLOCK_BOX_Y 20
#define TILE_KEY_PASS 54
#define TILE_WHITE_KEY 55
#define TILE_CHECKERED_KEY 56
#define TILE_BLACK_KEY 57
#define TILE_AMMO_CLIP 58
#define TILE_10_CREDITS 59
#define TILE_100_CREDITS 60
#define TILE_FIRST_AID 61
#define TILE_INVINCIBILITY 62
#define TILE_EXTRA_LIFE 63
#define TILE_UNOPEN_ALIEN_HOLE 69
#define TILE_OPEN_ALIEN_HOLE 70
#define TILE_WORM 71
#define TILE_FACE_HUGGER 72
#define TILE_NORMAL_ALIEN 73
#define TILE_STRONG_ALIEN 74
#define TILE_SECURITY_GUARD 75
#define TILE_WALL 89
#define TILE_VDOOR 168
#define TILE_HDOOR 169
#define TILE_VSDOOR 170
#define TILE_HSDOOR 171
#define TILE_VCWDOOR 172

/////////////////////////////////////////////////
// Typedefs
/////////////////////////////////////////////////
typedef struct
{
  unsigned char height;
  unsigned char width;
  unsigned char objectives;
  unsigned char time;
  unsigned char blockPutX;
  unsigned char blockPutY;
  unsigned char blockTakeX;
  unsigned char blockTakeY;
  unsigned char blockPutType;
  unsigned char blockTakeType;
  unsigned char type;
  unsigned char alienType;
  unsigned char finishX;
  unsigned char finishY;
  unsigned char startX;
  unsigned char startY;
  unsigned char data[MAX_HEIGHT][MAX_WIDTH];
} levelStructure;

/////////////////////////////////////////////////
// Global Variables
/////////////////////////////////////////////////
unsigned char numLevels, numTiles;
unsigned short hiScore;
levelStructure level[MAX_LEVELS];
unsigned char tile[256][8];
char levelsetName[13];
unsigned char passcodes[7][6];
unsigned char randomWall[] = {89, 89, 89, 89, 90, 91, 92, 93, 94, 95, 96, 97};
unsigned char randomInside[] = {10, 11, 12, 13};
unsigned char randomOutside[] = {7, 8, 9};


/////////////////////////////////////////////////
// Main Function
/////////////////////////////////////////////////
int main(int argc, char *argv[])
{
  FILE *fp, *listP;
  char listName[10];
  int quit = 0, i, j;
  char detect[DETECT_LENGTH];
  char *listLevelNames[] = {"level1.abc\n", "level2.abc\n", "level3.abc\n", "level4.abc\n", "level5.abc\n", "level6.abc\n", "level7.abc\n", "level8.abc\n"};
  char *listLevelNames2[] = {"level1.abc", "level2.abc", "level3.abc", "level4.abc", "level5.abc", "level6.abc", "level7.abc", "level8.abc"};

  randomize();

  printf("Alien Breed IV: Final Assault\n");
  printf("Levelset Editor v1.0\n\n");
  printf("Copyright (c)2001 TI-Calculator Programming Alliance\n");
  printf("Written and Programmed by James Vernon <james@calc.org>\n");
  printf("ICQ#: 71589304\n");
  printf("http://jamesv.calc.org\n");
  printf("http://tcpa.calc.org\n\n");

  fp = fopen("tiles.ab4", "rb");
  if (!fp)
  {
    printf("Couldn't find file tiles.ab4!\n");
    return ERR_TILES;
  }

  fread(&numTiles, 1, 1, fp);
  for (i = 0 ; i < (int)numTiles ; i++)
    fread(&tile[i], 8, 1, fp);
  fclose(fp);

  if (argc != 2)
  {
    printf("USAGE: ab4edit <levelset>\n");
    return ERR_USAGE;
  }

  if (strlen(argv[1]) != 3)
  {
    printf("Invalid Levelset name: \"%s\"!\n", argv[1]);
    return ERR_INVALIDNAME;
  }

  strcpy(levelsetName, "zab4l");
  strcat(levelsetName, argv[1]);
  strcat(levelsetName, ".ab4");

  fp = fopen(levelsetName, "rb");
  if (!fp)
    createNew();
  else
  {
    loadLevelset(fp);
    fclose(fp);
  }

  while (1)
  {
    clrscr();
    printf("\n");
    printf("ͻ");
    printf("             Alien Breed IV: Final Assault - Levelset Editor v1.0             ");
    printf("͹");
    printf(" Levelset: %3s                                                    Levels: %d ", strupr(argv[1]), (int)numLevels);
    printf("ͼ\n\n");
    printf("1.  Edit Levels\n");
    printf("2.  Help\n");
    printf("3.  Save & Exit\n");
    printf("4.  Exit (no save)\n\n");
    printf("Enter your selection: ");
    switch (getch())
    {
      case K_1:
        editLevels();
        break;
      case K_2:
        helpScreen();
        break;
      case K_3:
        quit = 1;
        break;
      case K_4:
        quit = 2;
        break;
    }
    if (quit && !checkAHandSG())
      break;
    if (quit && checkAHandSG())
    {
      printf("\n\nEither too many Alien Holes or Security Guards on Level %d!", checkAHandSG());
      waitKey();
      quit = 0;
    }
  }

  clrscr();
  printf("Alien Breed IV: Final Assault\n");
  printf("Levelset Editor v1.0\n\n");
  printf("Copyright (c)2001 TI-Calculator Programming Alliance\n");
  printf("Written and Programmed by James Vernon <james@calc.org>\n");
  printf("ICQ#: 71589304\n");
  printf("http://jamesv.calc.org\n");
  printf("http://tcpa.calc.org\n\n");

  if (quit == 1)
    saveLevelset(levelsetName);

  fp = fopen(levelsetName, "rb");
  if (!fp)
    return 0;
  loadLevelset(fp);

  listName[0] = NULL;
  strcat(listName, argv[1]);
  strcat(listName, ".lst");

  listP = fopen(listName, "w");
  fp = fopen("header.abc", "wb");
  strcpy(detect, "AB40");
  strncat(detect, strupr(argv[1]), 3);
  fwrite(detect, DETECT_LENGTH - 1, 1, fp);
  fwrite(&numLevels, 1, 1, fp);
  fwrite(&hiScore, 2, 1, fp);
  fwrite(passcodes, 6, 7, fp);
  fclose(fp);

  for (i = 0 ; i < numLevels ; i++)
  {
    fprintf(listP, listLevelNames[i]);
    fp = fopen(listLevelNames2[i], "wb");
    fwrite(&level[i], sizeof(level[i]) - (MAX_HEIGHT * MAX_WIDTH), 1, fp);
    for (j = 0 ; j < level[i].height ; j++)
      fwrite(level[i].data[j], (int)level[i].width, 1, fp);
    fclose(fp);
  }

  return NO_ERR;
}

/////////////////////////////////////////////////
// createNew - Initialise all Levelset variables
//
// Input:    void
// Output:   void
/////////////////////////////////////////////////
void createNew(void)
{
  int i, j, k;

  numLevels = 1;
  hiScore = 0;
  for (i = 0 ; i < 7 ; i++)
    strcpy(passcodes[i], "\0\0\0\0\0\0");
  for (i = 0 ; i < MAX_LEVELS ; i++)
  {
    level[i].height = MIN_HEIGHT;
    level[i].width = MIN_WIDTH;
    level[i].objectives = 0;
    level[i].time = 99;
    level[i].blockPutX = 1;
    level[i].blockPutY = 0;
    level[i].blockTakeX = 2;
    level[i].blockTakeY = 0;
    level[i].blockPutType = TILE_WALL;
    level[i].blockTakeType = 0;
    level[i].type = 0;
    level[i].alienType = 3;
    level[i].finishX = MIN_WIDTH - 1;
    level[i].finishY = MIN_HEIGHT - 1;
    level[i].startX = 0;
    level[i].startY = 0;
    for (j = 0 ; j < MAX_HEIGHT ; j++)
      for (k = 0 ; k < MAX_WIDTH ; k++)
        level[i].data[j][k] = 0;
  }
}

/////////////////////////////////////////////////
// loadLevelset - Try to load a Levelset
//
// Input:    FILE => Start of file
// Output:   void
/////////////////////////////////////////////////
void loadLevelset(FILE *fp)
{
  char detect[DETECT_LENGTH];
  int i, j;

  createNew();

  fread(detect, DETECT_LENGTH - 1, 1, fp);
  fread(&numLevels, 1, 1, fp);
  fread(&hiScore, 2, 1, fp);
  fread(passcodes, 6, 7, fp);
  for (i = 0 ; i < MAX_LEVELS ; i++)
  {
    fread(&level[i], sizeof(level[i]) - (MAX_HEIGHT * MAX_WIDTH), 1, fp);
    for (j = 0 ; j < level[i].height ; j++)
      fread(level[i].data[j], (int)level[i].width, 1, fp);
  }
  fclose(fp);
}

/////////////////////////////////////////////////
// saveLevelset - Save Levelset
//
// Input:    char => Name of file to save to
// Output:   void
/////////////////////////////////////////////////
void saveLevelset(char *name)
{
  FILE *fp;
  char detect[DETECT_LENGTH];
  int i, j;

  fp = fopen(name, "wb");

  strcpy(detect, "AB40");
  strncat(detect, strupr(name)+5, 3);

  fwrite(detect, DETECT_LENGTH - 1, 1, fp);
  fwrite(&numLevels, 1, 1, fp);
  fwrite(&hiScore, 2, 1, fp);
  fwrite(passcodes, 6, 7, fp);
  for (i = 0 ; i < numLevels ; i++)
  {
    fwrite(&level[i], sizeof(level[i]) - (MAX_HEIGHT * MAX_WIDTH), 1, fp);
    for (j = 0 ; j < level[i].height ; j++)
      fwrite(level[i].data[j], (int)level[i].width, 1, fp);
  }
  fclose(fp);
}

/////////////////////////////////////////////////
// editLevels - Level Editor screen
//
// Input:    void
// Output:   void
/////////////////////////////////////////////////
void editLevels(void)
{
  int levelNo = 0,
      quit = 0,
      i, j, k, l, m;
  unsigned char tileNo = 0,
                xOffset = 0,
                yOffset = 0,
                x = 0,
                y = 0;

  gfxMode();
  showTiles();

  while (1)
  {
    setCursor(0, 0);
    printf("Level:    %2d/%2d\n", levelNo + 1, (int)numLevels);
    printf("X:        %2d/%2d\n", (int)(x + xOffset), (int)(level[levelNo].width-1));
    printf("Y:        %2d/%2d\n", (int)(y + yOffset), (int)(level[levelNo].height - 1));
    printf("Type:     ");
    if (!level[levelNo].type)
      printf("Normal  \n");
    else
      printf("Hatchery\n");
    printf("Passcode: ");
    if (levelNo)
    {
      if (!passcodes[levelNo - 1][0])
        printf("None \n");
      else
        printf("%s\n", passcodes[levelNo - 1]);
    }
    else
      printf("None \n");
    printf("Time:     %3d\n", (int)level[levelNo].time);
    printf("Hole Alien:\n       ");
    if (level[levelNo].alienType == 3)
      printf("Normal");
    else
      printf("Strong");
    setCursor(0, 20);
    printf("Selected Block: %3d", (int)level[levelNo].data[y + yOffset][x + xOffset]);
    setCursor(1, 20);
    printf("Selected Tile:  %3d", (int)tileNo);
    setCursor(5, 20);
    printf("Objectives:     %3d", (int)level[levelNo].objectives);
    showLevel(levelNo, xOffset, yOffset);
    highlightTile((int)tileNo, BLUE);
    highlightBlock((int)x, (int)y);
    if ((level[levelNo].startX >= xOffset) && (level[levelNo].startX <= (xOffset + 11)) && (level[levelNo].startY >= yOffset) && (level[levelNo].startY <= (yOffset + 7)))
      drawBox((level[levelNo].startY - yOffset) * 8 + VIEW_BOX_Y + 2, (level[levelNo].startX - xOffset) * 8 + VIEW_BOX_X + 2, 6, 6, LIGHTGRAY);
    if ((level[levelNo].finishX >= xOffset) && (level[levelNo].finishX <= (xOffset + 11)) && (level[levelNo].finishY >= yOffset) && (level[levelNo].finishY <= (yOffset + 7)))
      drawBox((level[levelNo].finishY - yOffset) * 8 + VIEW_BOX_Y + 2, (level[levelNo].finishX - xOffset) * 8 + VIEW_BOX_X + 2, 6, 6, DARKGRAY);
    if ((level[levelNo].blockPutX >= xOffset) && (level[levelNo].blockPutX <= (xOffset + 11)) && (level[levelNo].blockPutY >= yOffset) && (level[levelNo].blockPutY <= (yOffset + 7)))
      drawBox((level[levelNo].blockPutY - yOffset) * 8 + VIEW_BOX_Y + 2, (level[levelNo].blockPutX - xOffset) * 8 + VIEW_BOX_X + 2, 6, 6, YELLOW);
    if ((level[levelNo].blockTakeX >= xOffset) && (level[levelNo].blockTakeX <= (xOffset + 11)) && (level[levelNo].blockTakeY >= yOffset) && (level[levelNo].blockTakeY <= (yOffset + 7)))
      drawBox((level[levelNo].blockTakeY - yOffset) * 8 + VIEW_BOX_Y + 2, (level[levelNo].blockTakeX - xOffset) * 8 + VIEW_BOX_X + 2, 6, 6, BROWN);
    drawBox(BLOCK_BOX_Y, BLOCK_BOX_X, 10, 10, YELLOW);
    drawBox(BLOCK_BOX_Y, BLOCK_BOX_X + 12, 10, 10, BROWN);
    putBWSprite(BLOCK_BOX_X + 1, BLOCK_BOX_Y + 1, tile[level[levelNo].blockPutType]);
    putBWSprite(BLOCK_BOX_X + 13, BLOCK_BOX_Y + 1, tile[level[levelNo].blockTakeType]);
    switch (getch())
    {
      case K_ESC:
        quit = 1;
        break;
      case K_SPACE:
        level[levelNo].data[y + yOffset][x + xOffset] = tileNo;
        break;
      case K_TAB:
        level[levelNo].type ^= 0x01;
        break;
      case K_ENTER:
        for (i = 0 ; i < MAX_HEIGHT ; i++)
          for (j = 0 ; j < MAX_WIDTH ; j++)
            switch (level[levelNo].type)
            {
              case 0:
                if (level[levelNo].data[i][j] == 1)
                  level[levelNo].data[i][j] = 0;
                break;
              case 1:
                if (level[levelNo].data[i][j] == 0)
                  level[levelNo].data[i][j] = 1;
                break;
            }
        break;
      case K_CTRLA:
        if (level[levelNo].alienType == 3)
          level[levelNo].alienType = 4;
        else
          level[levelNo].alienType = 3;
        break;
      case K_CTRLS:
        saveLevelset(levelsetName);
        setCursor(13, 0);
        printf("Saved :)");
        waitKey();
        setCursor(13, 0);
        printf("        ");
        break;
      case K_1:
        level[levelNo].data[y + yOffset][x + xOffset] = randomWall[random(sizeof(randomWall) / sizeof(unsigned char))];
        break;
      case K_2:
        level[levelNo].data[y + yOffset][x + xOffset] = randomInside[random(sizeof(randomInside) / sizeof(unsigned char))];
        break;
      case K_3:
        level[levelNo].data[y + yOffset][x + xOffset] = randomOutside[random(sizeof(randomOutside) / sizeof(unsigned char))];
        break;
      case K_5:
        for (i = 0, k = 0 ; i < level[levelNo].height ; i++)
          for (j = 0 ; j < level[levelNo].width ; j++)
            if ((level[levelNo].data[i][j] == TILE_UNOPEN_ALIEN_HOLE) || (level[levelNo].data[i][j] == TILE_OPEN_ALIEN_HOLE))
              k++;
        setCursor(13, 0);
        printf("Holes: %d", k);
        waitKey();
        setCursor(13, 0);
        printf("         ");
        break;
      case K_6:
        for (i = 0, k = 0 ; i < level[levelNo].height ; i++)
          for (j = 0 ; j < level[levelNo].width ; j++)
            if (level[levelNo].data[i][j] == TILE_WORM)
              k++;
        setCursor(13, 0);
        printf("Worms: %d", k);
        waitKey();
        setCursor(13, 0);
        printf("         ");
        break;
      case K_7:
        for (i = 0, k = 0 ; i < level[levelNo].height ; i++)
          for (j = 0 ; j < level[levelNo].width ; j++)
            if (level[levelNo].data[i][j] == TILE_FACE_HUGGER)
              k++;
        setCursor(12, 0);
        printf("Face Huggers:\n");
        printf(" %d", k);
        waitKey();
        setCursor(12, 0);
        printf("             \n");
        printf("   ");
        break;
      case K_8:
        for (i = 0, k = 0 ; i < level[levelNo].height ; i++)
          for (j = 0 ; j < level[levelNo].width ; j++)
            if (level[levelNo].data[i][j] == TILE_NORMAL_ALIEN)
              k++;
        setCursor(12, 0);
        printf("Normal Aliens:\n");
        printf(" %d", k);
        waitKey();
        setCursor(12, 0);
        printf("              \n");
        printf("   ");
        break;
      case K_9:
        for (i = 0, k = 0 ; i < level[levelNo].height ; i++)
          for (j = 0 ; j < level[levelNo].width ; j++)
            if (level[levelNo].data[i][j] == TILE_STRONG_ALIEN)
              k++;
        setCursor(12, 0);
        printf("Strong Aliens:\n");
        printf(" %d", k);
        waitKey();
        setCursor(12, 0);
        printf("              \n");
        printf("   ");
        break;
      case K_0:
        for (i = 0, k = 0 ; i < level[levelNo].height ; i++)
          for (j = 0 ; j < level[levelNo].width ; j++)
            if (level[levelNo].data[i][j] == TILE_SECURITY_GUARD)
              k++;
        setCursor(12, 0);
        printf("Security Guards:\n");
        printf(" %d", k);
        waitKey();
        setCursor(12, 0);
        printf("                \n");
        printf("   ");
        break;
      case K_Q:
        for (i = 0, k = 0 ; i < level[levelNo].height ; i++)
          for (j = 0 ; j < level[levelNo].width ; j++)
          {
            if (level[levelNo].data[i][j] == TILE_10_CREDITS)
              k += 10;
            if (level[levelNo].data[i][j] == TILE_100_CREDITS)
              k += 100;
          }
        setCursor(12, 0);
        printf("Credits:\n");
        printf(" %d", k);
        waitKey();
        setCursor(12, 0);
        printf("        \n");
        printf("      ");
        break;
      case K_W:
        for (i = 0, k = 0 ; i < level[levelNo].height ; i++)
          for (j = 0 ; j < level[levelNo].width ; j++)
            if (level[levelNo].data[i][j] == TILE_AMMO_CLIP)
              k++;
        setCursor(12, 0);
        printf("Ammo Clips: %d\n", k);
        printf("Bullets: %d", k * 20);
        waitKey();
        setCursor(12, 0);
        printf("              \n");
        printf("             ");
        break;
      case K_E:
        for (i = 0, k = 0 ; i < level[levelNo].height ; i++)
          for (j = 0 ; j < level[levelNo].width ; j++)
            if (level[levelNo].data[i][j] == TILE_KEY_PASS)
              k++;
        setCursor(12, 0);
        printf("Key Passes:\n");
        printf(" %d", k);
        waitKey();
        setCursor(12, 0);
        printf("           \n");
        printf("   ");
        break;
      case K_R:
        for (i = 0, k = 0, l = 0, m = 0 ; i < level[levelNo].height ; i++)
          for (j = 0 ; j < level[levelNo].width ; j++)
          {
            if (level[levelNo].data[i][j] == TILE_WHITE_KEY)
              k++;
            if (level[levelNo].data[i][j] == TILE_CHECKERED_KEY)
              l++;
            if (level[levelNo].data[i][j] == TILE_BLACK_KEY)
              m++;
          }
        setCursor(11, 0);
        printf("Coloured Keys:\n");
        printf("W     C     B\n");
        if (k)
          printf("Yes   ");
        else
          printf("No    ");
        if (l)
          printf("Yes   ");
        else
          printf("No    ");
        if (m)
          printf("Yes");
        else
          printf("No");
        waitKey();
        setCursor(11, 0);
        printf("              \n");
        printf("             \n");
        printf("               ");
        break;
      case K_T:
        for (i = 0, k = 0 ; i < level[levelNo].height ; i++)
          for (j = 0 ; j < level[levelNo].width ; j++)
            if (level[levelNo].data[i][j] == TILE_FIRST_AID)
              k++;
        setCursor(12, 0);
        printf("First Aids:\n");
        printf(" %d", k);
        waitKey();
        setCursor(12, 0);
        printf("           \n");
        printf("   ");
        break;
      case K_Y:
        for (i = 0, k = 0 ; i < level[levelNo].height ; i++)
          for (j = 0 ; j < level[levelNo].width ; j++)
            if (level[levelNo].data[i][j] == TILE_INVINCIBILITY)
              k++;
        setCursor(12, 0);
        printf("Invincibilities:\n");
        printf(" %d", k);
        waitKey();
        setCursor(12, 0);
        printf("                \n");
        printf("   ");
        break;
      case K_U:
        for (i = 0, k = 0 ; i < level[levelNo].height ; i++)
          for (j = 0 ; j < level[levelNo].width ; j++)
            if (level[levelNo].data[i][j] == TILE_EXTRA_LIFE)
              k++;
        setCursor(12, 0);
        printf("Extra Lives:\n");
        printf(" %d", k);
        waitKey();
        setCursor(12, 0);
        printf("            \n");
        printf("   ");
        break;
      case K_I:
        for (i = 0, k = 0 ; i < level[levelNo].height ; i++)
          for (j = 0 ; j < level[levelNo].width ; j++)
            if ((level[levelNo].data[i][j] == TILE_VDOOR) || (level[levelNo].data[i][j] == TILE_HDOOR))
              k++;
        setCursor(13, 0);
        printf("Doors: %d", k);
        waitKey();
        setCursor(13, 0);
        printf("         ");
        break;
      case K_O:
        for (i = 0, k = 0 ; i < level[levelNo].height ; i++)
          for (j = 0 ; j < level[levelNo].width ; j++)
            if ((level[levelNo].data[i][j] == TILE_VSDOOR) || (level[levelNo].data[i][j] == TILE_HSDOOR))
              k++;
        setCursor(12, 0);
        printf("Security Doors:\n");
        printf(" %d", k);
        waitKey();
        setCursor(12, 0);
        printf("               \n");
        printf("    ");
        break;
      case K_P:
        for (i = 0, k = 0 ; i < level[levelNo].height ; i++)
          for (j = 0 ; j < level[levelNo].width ; j++)
            if ((level[levelNo].data[i][j] >= TILE_VCWDOOR) && (level[levelNo].data[i][j] <= (TILE_VCWDOOR + 5)))
              k++;
        setCursor(12, 0);
        printf("Coloured Doors:\n");
        printf(" %d", k);
        waitKey();
        setCursor(12, 0);
        printf("               \n");
        printf("    ");
        break;
      case 0:
        switch (getch())
        {
          case K_UP:
            if (y)
              y--;
            else
              if (yOffset)
                yOffset--;
            break;
          case K_DOWN:
            if (y < 7)
              y++;
            else
              if (yOffset < (level[levelNo].height - 8))
                yOffset++;
            break;
          case K_LEFT:
            if (x)
              x--;
            else
              if (xOffset)
                xOffset--;
            break;
          case K_RIGHT:
            if (x < 11)
              x++;
            else
              if (xOffset < (level[levelNo].width - 12))
                xOffset++;
            break;
          case K_HOME:
            highlightTile((int)tileNo, GREEN);
            tileNo -= 32;
            break;
          case K_END:
            highlightTile((int)tileNo, GREEN);
            tileNo += 32;
            break;
          case K_DELETE:
            highlightTile((int)tileNo, GREEN);
            tileNo--;
            break;
          case K_PGDOWN:
            highlightTile((int)tileNo, GREEN);
            tileNo++;
            break;
          case K_F1:
            if (level[levelNo].width > MIN_WIDTH)
              level[levelNo].width--;
            if ((xOffset + 12) > level[levelNo].width)
              xOffset--;
            break;
          case K_F2:
            if (level[levelNo].width < MAX_WIDTH)
              level[levelNo].width++;
            break;
          case K_F3:
            if (level[levelNo].height > MIN_HEIGHT)
              level[levelNo].height--;
            if ((yOffset + 8) > level[levelNo].height)
              yOffset--;
            break;
          case K_F4:
            if (level[levelNo].height < MAX_HEIGHT)
              level[levelNo].height++;
            break;
          case K_F5:
            if (numLevels == 1)
              break;
            if ((levelNo + 1) == numLevels)
              break;
            numLevels--;
            break;
          case K_F6:
            if (numLevels < MAX_LEVELS)
              numLevels++;
            break;
          case K_F7:
            if (levelNo > 0)
            {
              levelNo--;
              xOffset = yOffset = x = y = 0;
            }
            break;
          case K_F8:
            if ((levelNo < (MAX_LEVELS - 1)) && ((levelNo + 1) < numLevels))
            {
              levelNo++;
              xOffset = yOffset = x = y = 0;
            }
            break;
          case K_F9:
            level[levelNo].blockPutType = tileNo;
            break;
          case K_F10:
            level[levelNo].blockTakeType = tileNo;
            break;
          case K_CTRLF1:
            level[levelNo].startX = x + xOffset;
            level[levelNo].startY = y + yOffset;
            break;
          case K_CTRLF2:
            level[levelNo].finishX = x + xOffset;
            level[levelNo].finishY = y + yOffset;
            break;
          case K_CTRLF3:
            if (levelNo)
              setPasscode(levelNo - 1);
            break;
          case K_CTRLF4:
            showMap(levelNo);
            break;
          case K_CTRLF5:
            level[levelNo].time--;
            break;
          case K_CTRLF6:
            level[levelNo].time++;
            break;
          case K_CTRLF7:
            level[levelNo].objectives--;
            break;
          case K_CTRLF8:
            level[levelNo].objectives++;
            break;
          case K_CTRLF9:
            level[levelNo].blockPutX = x + xOffset;
            level[levelNo].blockPutY = y + yOffset;
            break;
          case K_CTRLF10:
            level[levelNo].blockTakeX = x + xOffset;
            level[levelNo].blockTakeY = y + yOffset;
            break;
        }
    }
    if (quit)
      break;
  }

  textMode();
}

/////////////////////////////////////////////////
// helpScreen - Display help screen
//
// Input:    void
// Output:   void
/////////////////////////////////////////////////
void helpScreen(void)
{
  clrscr();
  printf("HELP SCREEN - Keys for Editor Screen:\n\n");
  printf("ESC         - Return to Main Menu\n");
  printf("SPACEBAR    - Place tile at cursor\n");
  printf("CTRL+S      - Save Levelset\n");
  printf("CTRL+A      - Toggle alien type to come out of holes\n");
  printf("TAB         - Toggle level type\n");
  printf("ENTER       - Change floor to specified type\n");
  printf("Arrow Keys  - Move cursor around level\n");
  printf("HOME, DEL,\n");
  printf("END, PGDOWN - Move cursor around tiles\n\n");
  printf("Press any key for more...\n\n");
  waitKey();
  printf("F1          - Decrease level width\n");
  printf("F2          - Increase level width\n");
  printf("F3          - Decrease level height\n");
  printf("F4          - Increase level height\n");
  printf("F5          - Decrease number of levels\n");
  printf("F6          - Increase number of levels\n");
  printf("F7          - Move to previous level\n");
  printf("F8          - Move to next level\n");
  printf("F9          - Set current tile to BlockPutType\n");
  printf("F10         - Set current tile to BlockTakeType\n");
  printf("CTRL+F1     - Set start position to cursor position\n");
  printf("CTRL+F2     - Set finish position to cursor position\n");
  printf("CTRL+F3     - Change passcode for current level\n");
  printf("CTRL+F4     - Show map of entire level\n");
  printf("CTRL+F5     - Decrease time\n");
  printf("CTRL+F6     - Increase time\n");
  printf("CTRL+F7     - Decrease objectives\n");
  printf("CTRL+F8     - Increase objectives\n");
  printf("CTRL+F9     - Set BlockPut position to cursor position\n");
  printf("CTRL+F10    - Set BlockTake position to cursor position\n");
  printf("Press any key for more...\n\n");
  waitKey();
  printf("1           - Place random wall tile\n");
  printf("2           - Place random inside tile\n");
  printf("3           - Place random outside tile\n");
  printf("5           - Count Alien Holes on current level\n");
  printf("6           - Count Worms on current level\n");
  printf("7           - Count Face Huggers on current level\n");
  printf("8           - Count Normal Aliens on current level\n");
  printf("9           - Count Strong Aliens on current level\n");
  printf("0           - Count Security Guards on current level\n");
  printf("Q           - Count Credits on current level\n");
  printf("W           - Count Ammo Clips on current level\n");
  printf("E           - Count Key Passes on current level\n");
  printf("R           - Check Coloured Key Passes on current level\n");
  printf("T           - Count First Aids on current level\n");
  printf("Y           - Count Invincibilities on current level\n");
  printf("U           - Count Extra Lives on current level\n");
  printf("I           - Count Normal Doors on current level\n");
  printf("O           - Count Security Doors on current level\n");
  printf("P           - Count Coloured Doors on current level\n");
  waitKey();
}

/////////////////////////////////////////////////
// showMap - Show map view of level
//
// Input:    int = Level to show
// Output:   void
/////////////////////////////////////////////////
void showMap(int levelNo)
{
  int row, col, quit = 0, y = 0;

  while (1)
  {
    for (row = y ; row < (25 + y) ; row++)
      for (col = 0 ; col < MAX_WIDTH ; col++)
      {
        if (row >= level[levelNo].height)
          putBWSprite(col * 8 + 32, (row - y) * 8, tile[2]);
        else if (col >= level[levelNo].width)
          putBWSprite(col * 8 + 32, (row - y) * 8, tile[2]);
        else
          putBWSprite(col * 8 + 32, (row - y) * 8, tile[level[levelNo].data[row][col]]);
      }
    for (row = 0 ; row < 200 ; row += 8)
      for (col = 0 ; col < 32 ; col += 8)
      {
        putBWSprite(col, row, tile[2]);
        putBWSprite(col + 288, row, tile[2]);
      }
    switch (getch())
    {
      case K_ESC:
        quit = 1;
        break;
      case 0:
        switch (getch())
        {
          case K_UP:
            if (y)
              y--;
            break;
          case K_DOWN:
            if (y < 7)
              y++;
            break;
        }
        break;
    }
    if (quit)
      break;
  }
  clearVideo();
  showTiles();
}

/////////////////////////////////////////////////
// setPasscode - Get a new passcode for a level
//
// Input:    int = Level to set passcode for
// Output:   void
/////////////////////////////////////////////////
void setPasscode(int levelNo)
{
  char str[81];
  int i, j;

  textMode();
  gotoxy(1, 2);
  printf("(5 chars. in length, blank for none)");
  gotoxy(1, 1);
  printf("Enter new passcode: ");
  gets(str);
  if (!strlen(str))
    strcpy(passcodes[levelNo], "\0\0\0\0\0\0");
  else
  {
    if (strlen(str) == 5)
    {
      strcpy(str, strupr(str));
      for (i = 0, j = 0 ; i < 5 ; i++)
        if (!((str[i] >= 'A') && (str[i] <= 'Z')))
          j++;
      if (!j)
        strcpy(passcodes[levelNo], str);
      else
      {
        gotoxy(1, 5);
        printf("Enter letters only!");
        waitKey();
      }
    }
    else
    {
      gotoxy(1, 5);
      printf("Must be 5 chars. in length!");
      waitKey();
    }
  }
  gfxMode();
  showTiles();
}

/////////////////////////////////////////////////
// highlightBlock - Highlight selected Block in Level
//
// Input:    int = X on screen
//           int = Y on screen
// Output:   void
/////////////////////////////////////////////////
void highlightBlock(int x, int y)
{
  drawBox(y * 8 + VIEW_BOX_Y + 1, x * 8 + VIEW_BOX_X + 1, 8, 8, BLUE);
}

/////////////////////////////////////////////////
// highlightTile - Highlight selected Tile in Tile palette
//
// Input:    int = Tile to highlight
//           int = Color to highlight with
// Output:   void
/////////////////////////////////////////////////
void highlightTile(int tile, int color)
{
  drawBox(tile / 32 * 10 + SHOW_TILES_Y, tile % 32 * 10, 10, 10, color);
}

/////////////////////////////////////////////////
// showTiles - Show all tiles at bottom of screen
//
// Input:    void
// Output:   void
/////////////////////////////////////////////////
void showTiles(void)
{
  int i, x, y;

  fillBox(SHOW_TILES_Y, 0, 80, 320, GREEN);
  for (i = 0, x = 1, y = 121 ; i < (int)numTiles ; i++)
  {
    putBWSprite(x, y, tile[i]);
    x += 10;
    if (x > 320)
    {
      x = 1;
      y += 10;
    }
  }
}

/////////////////////////////////////////////////
// showLevel - Show a level
//
// Input:    int = Level to show
//           int = X Offset
//           int = Y Offset
// Output:   void
/////////////////////////////////////////////////
void showLevel(int l, int xOffset, int yOffset)
{
  int row, col;

  line(VIEW_BOX_X, VIEW_BOX_Y, VIEW_BOX_X + 97, VIEW_BOX_Y, RED);
  line(VIEW_BOX_X, VIEW_BOX_Y + 65, VIEW_BOX_X + 97, VIEW_BOX_Y + 65, RED);
  line(VIEW_BOX_X, VIEW_BOX_Y, VIEW_BOX_X, VIEW_BOX_Y + 65, RED);
  line(VIEW_BOX_X + 97, VIEW_BOX_Y, VIEW_BOX_X + 97, VIEW_BOX_Y + 65, RED);
  for (row = 0 ; row < 8 ; row++)
    for (col = 0 ; col < 12 ; col++)
      putBWSprite((col * 8) + VIEW_BOX_X + 1, (row * 8) + VIEW_BOX_Y + 1, tile[level[l].data[row + yOffset][col + xOffset]]);
}

/////////////////////////////////////////////////
// checkAHandSG - Check Levelset to make sure no levels have too many alien holes or security guards
//
// Input:    void
// Output:   0 = Check successful, otherwise, equals level that error occurred on
/////////////////////////////////////////////////
int checkAHandSG(void)
{
  int l, r, c;
  int h, s;

  for (l = 0 ; l < numLevels ; l++)
    for (r = 0, h = 0, s = 0 ; r < level[l].height ; r++)
      for (c = 0 ; c < level[l].width ; c++)
      {
        if ((level[l].data[r][c] == TILE_UNOPEN_ALIEN_HOLE) || (level[l].data[r][c] == TILE_OPEN_ALIEN_HOLE))
          h++;
        else if (level[l].data[r][c] == TILE_SECURITY_GUARD)
          s++;
        if ((h > 10) || (s > 5))
          return l + 1;
      }
  return 0;
}

/////////////////////////////////////////////////
// waitKey - Wait for a key press
//
// Input:    void
// Output:   void
/////////////////////////////////////////////////
void waitKey(void)
{
  while (!getch());
}

/////////////////////////////////////////////////
// gfxMode - Set video to mode 13h
//
// Input:    void
// Output:   void
/////////////////////////////////////////////////
void gfxMode(void)
{
  asm {
    mov ax,0x0013
    int 0x10
  }
}

/////////////////////////////////////////////////
// textMode - Set video to mode 03h
//
// Input:    void
// Output:   void
/////////////////////////////////////////////////
void textMode(void)
{
  asm {
    mov ax,0x0003
    int 0x10
  }
}

/////////////////////////////////////////////////
// setCursor - Set cursor coordinates
//
// Input:    unsigned char = Y
//           unsigned char = X
// Output:   void
/////////////////////////////////////////////////
void setCursor(unsigned char y, unsigned char x)
{
  asm {
    mov dh,y
    mov dl,x
    xor bh,bh
    mov ah,0x0002
    int 0x10
  }
}

/////////////////////////////////////////////////
// putPixel - Set a pixel on the screen
//
// Input:    int = X
//           int = Y
//           int = Color
// Output:   void
/////////////////////////////////////////////////
void putPixel(int x, int y, int c)
{
  unsigned short px, py;
  unsigned char pc;

  px = (unsigned short)x;
  py = (unsigned short)y;
  pc = (unsigned char)c;

  asm {
    mov di,0xA000
    mov es,di
    mov cx,py
    xor ax,ax
putPixelLoop:
    add ax,320
    loop putPixelLoop
    add ax,px
    mov di,ax
    mov al,pc
    mov [es:di],al
  }
}

/////////////////////////////////////////////////
// putBWSprite - Draw an 8x8 Black & White sprite to screen
//
// Input:    int = X
//           int = Y
//           unsigned char => Sprite data
// Output:   void
/////////////////////////////////////////////////
void putBWSprite(int x, int y, unsigned char *spr)
{
  int row, col, color;
  unsigned char bitmask;

  for (row = 0 ; row < 8 ; row++, spr++)
  {
    bitmask = 0x80;
    for (col = 0 ; col < 8 ; col++)
    {
      if (!(*spr & bitmask))
        color = WHITE;
      else
        color = BLACK;
      putPixel(col + x, row + y, color);
      bitmask >>= 1;
    }
  }
}

/////////////////////////////////////////////////
// fillBox - Fill a box of any height/width
//
// Input:    int = Y
//           int = X
//           int = Height
//           int = Width
//           int = Color
// Output:   void
/////////////////////////////////////////////////
void fillBox(int y, int x, int h, int w, int c)
{
  int row, col;

  for (row = y ; row < (h + y) ; row++)
    for (col = x ; col < (w + x) ; col++)
      putPixel(col, row, (unsigned char)c);
}

/////////////////////////////////////////////////
// line - Draw a line
//
// Input:    unsigned short = X1
//           unsigned short = Y1
//           unsigned short = X2
//           unsigned short = Y2
//           unsigned char = Color
// Output:   void
/////////////////////////////////////////////////
void line(unsigned short x1, unsigned short y1, unsigned short x2, unsigned short y2, unsigned char c)
{
  short xi, yi, dx, dy, d, ai, bi;

  if (x1 < x2)
  {
    xi = 1;
    dx = x2 - x1;
  }
  else
  {
    xi = -1;
    dx = x1 - x2;
  }

  if (y1 < y2)
  {
    yi = 1;
    dy = y2 - y1;
  }
  else
  {
    yi = -1;
    dy = y1 - y2;
  }

  putPixel(x1, y1, c);
  if (dx > dy)
  {
    ai = (dy - dx) * 2;
    bi = dy * 2;
    d = bi - dx;
    do
    {
      if (d >= 0)
      {
        y1 += yi;
        d += ai;
      }
      else
        d += bi;
      x1 += xi;
      putPixel(x1, y1, c);
    } while (x1 != x2);
  }
  else
  {
    ai = (dx - dy) * 2;
    bi = dx * 2;
    d = bi - dy;
    do
    {
      if (d >= 0)
      {
        x1 += xi;
        d += ai;
      }
      else
        d += bi;
      y1 += yi;
      putPixel(x1, y1, c);
    } while (y1 != y2);
  }
}

/////////////////////////////////////////////////
// drawBox - Draw a box of any height/width
//
// Input:    int = Y
//           int = X
//           int = Height
//           int = Width
//           int = Color
// Output:   void
/////////////////////////////////////////////////
void drawBox(int y, int x, int h, int w, int c)
{
  line((unsigned short)x, (unsigned short)y, (unsigned short)(x + w - 1), (unsigned short)y, (unsigned char)c);
  line((unsigned short)x, (unsigned short)(y + h - 1), (unsigned short)(x + w - 1), (unsigned short)(y + h - 1), (unsigned char)c);
  line((unsigned short)x, (unsigned short)y, (unsigned short)x, (unsigned short)(y + h - 1), (unsigned char)c);
  line((unsigned short)(x + w - 1), (unsigned short)y, (unsigned short)(x + w - 1), (unsigned short)(y + h - 1), (unsigned char)c);
}

/////////////////////////////////////////////////
// clearVideo - Clear video screen
//
// Input:    void
// Output:   void
/////////////////////////////////////////////////
void clearVideo(void)
{
  asm {
    mov di,0xA000
    mov es,di
    mov cx,320*200
    mov di,0x0000
    xor al,al
clearVideoLoop:
    mov [es:di],al
    inc di
    loop clearVideoLoop
  }
}
